from .arena_object import Object

class Dodecahedron(Object):
    """
    Class for Dodecahedron in the ARENA.
    """
    def __init__(self, **kwargs):
        super().__init__(object_type="dodecahedron", **kwargs)



from ..utils import Utils
from .attribute import Attribute
from scipy.spatial.transform import Rotation as R

class Rotation(Attribute):
    """
    Rotation Attribute in quaternions or euler.
    Usage: rotation=Rotation(x,y,z,w) or rotation=Rotation(x,y,z)
    """
    def __init__(self, x=0, y=0, z=0, w=None):
        if w is not None: # quaternions
            super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z), w=Utils.agran(w))
        else: # euler
            super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z), w=None)

    def __repr__(self):
        if self.is_quaternion:
            return str(vars(self))
        else:
            rot_dict = vars(self).copy()
            del rot_dict["w"]
            return str(rot_dict)

    @property
    def is_quaternion(self):
        return self.w is not None

    @property
    def euler(self):
        if self.is_quaternion: # quaternions
            euler = Rotation.q2e((self.x, self.y, self.z, self.w))
            return Rotation(*euler)
        else: # euler
            return self

    @property
    def quaternion(self):
        if self.is_quaternion: # quaternions
            return self
        else: # euler
            quat = Rotation.e2q((self.x, self.y, self.z))
            return Rotation(*quat)

    @classmethod
    def q2e(cls, q):
        """quaternions to euler"""
        rot = R.from_quat(q)
        return rot.as_euler('xyz', degrees=False)

    @classmethod
    def e2q(cls, e):
        """euler to quaternions"""
        rot = R.from_euler('xyz', e, degrees=False)
        return rot.as_quat()



from ..utils import Utils
from .attribute import Attribute

class Scale(Attribute):
    """
    Scale Attribute.
    Usage: scale=Scale(x,y,z)
    """
    def __init__(self, x=1, y=1, z=1):
        super().__init__(x=Utils.agran(x), y=Utils.agran(y), z=Utils.agran(z))

    def to_str(self):
        return Utils.tuple_to_string((self.x, self.y, self.z))



from .arena_object import Object

class Text(Object):
    """
    Class for Text in the ARENA.
    [TODO]: update_attribute(text="new text") does get published. maybe a client-side issue?
    """
    def __init__(self, text="placeholder text", **kwargs):
        super().__init__(object_type="text", text=text, **kwargs)



from .attribute import Attribute

class Color(Attribute):
    """
    Color Attribute.
    Usage: color=Color(red,green,blue)
    """
    def __init__(self, red=0, green=0, blue=0):
        super().__init__(red=red, green=green, blue=blue)

    @property
    def hex(self):
        return "#{:02x}{:02x}{:02x}".format(self.red, self.green, self.blue)



from .attribute import Attribute
from .position import Position

class Physics(Attribute):
    """
    Physics Attribute.
    Usage: physics=Physics(...) OR dynamic_body=Physics(...)
    """
    def __init__(self, type="static"):
        _type = type
        if _type != "static" and _type != "dynamic":
            _type = "none"
        super().__init__(type=_type)



from ..utils import Utils
from .attribute import Attribute

class Sound(Attribute):
    """
    Sound Attribute.
    Usage: sound=Sound(...)
    """
    def __init__(self, src, positional=False, **kwargs):
        super().__init__(src=src, positional=positional, **kwargs)



from .arena_object import Object

class Box(Object):
    """
    Class for Box in the ARENA.
    """
    def __init__(self, **kwargs):
        super().__init__(object_type="box", **kwargs)

class Cube(Object):
    """
    Class for Cube in the ARENA.
    """
    def __init__(self, **kwargs):
        print("[WARNING]", "Cube will be deprecated soon, please use Box instead!")
        super().__init__(object_type="cube", **kwargs)



from ..base_object import *
from ..attributes import Data
from ..utils import *
import uuid


class Object(BaseObject):
    """
    Object class. Defines a generic object in the ARENA.
    """

    all_objects = {} # dict of all objects created so far

    def __init__(self, evt_handler=None, update_handler=None, **kwargs):
        # "object_id" is required in kwargs, defaulted to random uuid4
        object_id = kwargs.get("object_id", str(uuid.uuid4()))
        if "object_id" in kwargs: del kwargs["object_id"]

        # "persist" is required in kwargs, defaulted to false
        persist = kwargs.get("persist", False)
        if "persist" in kwargs: del kwargs["persist"]

        # special case for "parent" (can be an Object)
        if "parent" in kwargs and isinstance(kwargs["parent"], Object):
            kwargs["parent"] = kwargs["parent"].object_id

        # "ttl" is optional
        ttl = kwargs.get("ttl", None)
        if "ttl" in kwargs: del kwargs["ttl"]

        # remove timestamp, if exists
        if "timestamp" in kwargs: del kwargs["timestamp"]

        # remove "action", if exists
        if "action" in kwargs: del kwargs["action"]

        # print warning if object is being created with the same id as an existing object
        if Object.exists(object_id):
            print("[WARNING]", f"An object with object_id of {object_id} was already created. The previous object will be overwritten.")
            Object.remove(Object.get(object_id))

        # setup attributes in the "data" field
        data = kwargs.get("data", kwargs)
        data = Data(**data)
        if ttl:
            super().__init__(
                    object_id=object_id,
                    type="object",
                    persist=persist,
                    ttl=ttl,
                    data=data
                )
        else:
            super().__init__(
                    object_id=object_id,
                    type="object",
                    persist=persist,
                    data=data
                )

        self.evt_handler = evt_handler
        self.update_handler = update_handler

        # add current object to all_objects dict
        Object.add(self)

    def update_attributes(self, evt_handler=None, update_handler=None, **kwargs):
        if evt_handler:
            self.evt_handler = evt_handler

        if update_handler:
            self.update_handler = update_handler

        if "data" not in self:
            return

        # update "persist", and "ttl"
        self.persist = kwargs.get("persist", self.persist)
        if "ttl" in self:
            self.ttl = kwargs.get("ttl", self.ttl)

        data = self.data
        Data.update_data(data, kwargs)

        if self.update_handler:
            self.update_handler(self)

    def json(self, **kwargs):
        # kwargs are for additional param to add to json, like "action":"create"
        res = { k:v for k,v in vars(self).items() if k != "evt_handler" and k != "update_handler" }
        res.update(kwargs)

        data = res["data"].__dict__.copy()

        # color should be a hex string
        if "color" in data:
            data["color"] = data["color"].hex

        # rotation should be in quaternions
        if "rotation" in data:
            data["rotation"] = data["rotation"].quaternion

        # handle special case where "physics" should be "dynamic-body"
        if "physics" in data:
            ref = data["physics"]
            del data["physics"]
            data["dynamic-body"] = ref

        if "dynamic_body" in data:
            ref = data["dynamic_body"]
            del data["dynamic_body"]
            data["dynamic-body"] = ref

        # handle special case where "clickable" should be "click-listener"
        if "clickable" in data:
            ref = data["clickable"]
            del data["clickable"]
            data["click-listener"] = ref

        if "click_listener" in data:
            ref = data["click_listener"]
            del data["click_listener"]
            data["click-listener"] = ref

        # remove underscores from specific keys
        if "goto_url" in data:
            ref = data["goto_url"]
            del data["goto_url"]
            data["goto-url"] = ref

        if "animation_mixer" in data:
            ref = data["animation_mixer"]
            del data["animation_mixer"]
            data["animation-mixer"] = ref

        # for animation, replace "start" and "end" with "from" and "to"
        if "animation" in data:
            animation = data["animation"].__dict__.copy()
            if "start" in animation:
                animation["from"] = animation["start"]
                del animation["start"]
            if "end" in animation:
                animation["to"] = animation["end"]
                del animation["end"]
            data["animation"] = animation

        res["data"] = data
        return self.json_encode(res)

    # methods for global object dictionary
    @classmethod
    def get(cls, object_id):
        return Object.all_objects.get(object_id, None)

    @classmethod
    def add(cls, obj):
        object_id = obj.object_id
        Object.all_objects[object_id] = obj

    @classmethod
    def remove(cls, obj):
        object_id = obj.object_id
        del Object.all_objects[object_id]

    @classmethod
    def exists(cls, object_id):
        return object_id in Object.all_objects



from .arena_object import Object

class GLTF(Object):
    """
    Class for GLTF Model in the ARENA.
    """
    def __init__(self, url="", **kwargs):
        super().__init__(object_type="gltf-model", url=url, **kwargs)



from .attribute import Attribute

class GotoUrl(Attribute):
    """
    GotoUrl Attribute.
    Usage: goto_url=GotoUrl(...)
    """
    def __init__(self, url, dest="static", on="mousedown"):
        if dest != "popup" and dest != "newtab" and dest != "sametab":
            dest = "popup"
        super().__init__(url=url, dest=dest, on=on)



import json

class CustomEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for nested BaseObjects
    """
    def default(self, obj):
        if type(obj) == dict:
            return obj
        else:
            return vars(obj)

class BaseObject(object):
    """
    Basic Building Block for everything in ARENA-py.
    Can easily be interpreted and used like a JSON-able Python dictionary.
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return str(vars(self))

    def __getitem__(self, name):
        return self.__dict__[name]

    def __setitem__(self, name, attr):
        self.add(name, attr)

    def __contains__(self, attr):
        return attr in self.__dict__

    def add(self, name, attr):
        self.__dict__[name] = attr

    def json_encode(self, d):
        return json.dumps(d, cls=CustomEncoder)

    def json(self, **kwargs): # kwargs are for additional param to add to json, like "action":"create"
        res = vars(self).copy()
        res.update(kwargs)
        return self.json_encode(res)



import signal
import asyncio

class EventLoop(object):
    """
    Wrapper for an asyncio event loop.
    """
    def __init__(self, shutdown_func=None):
        self.tasks = []
        self.loop = asyncio.get_event_loop()
        self.signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        self.shutdown_func = shutdown_func

    async def _shutdown(self):
        if self.shutdown_func: self.shutdown_func()
        self.future.cancel()
        self.loop.stop()

    def add_task(self, worker):
        self.tasks += [worker.run()]

    def run(self):
        # cancellation of Future ensures all tasks (even if not completed) are cancelled
        self.future = asyncio.gather(*self.tasks)
        for s in self.signals:
            self.loop.add_signal_handler(
                s, lambda s=s: asyncio.ensure_future(self._shutdown())
            )

        # run event loop
        try:
            self.loop.run_forever()
        except asyncio.exceptions.CancelledError:
            pass
        finally:
            self.loop.close()



from .arena_object import Object
from ..attributes import Position, Rotation

class Camera(Object):
    """
    Class for Camera in the ARENA.
    """
    def __init__(self, object_id, **kwargs):
        data = kwargs.get("data", kwargs)

        self.hasAudio = kwargs.get("hasAudio", False)
        self.hasVideo = kwargs.get("hasVideo", False)
        self.hasAvatar = kwargs.get("hasAvatar", False)
        self.displayName = kwargs.get("displayName", "")
        self.jistsiId = kwargs.get("jistsiId", None)

        position = data.get("position", None)
        rotation = data.get("rotation", None)

        if position is not None and rotation is not None:
            super().__init__(object_type="camera", object_id=object_id, position=Position(**position), rotation=Rotation(**rotation), **kwargs)
        elif position is not None:
            super().__init__(object_type="camera", object_id=object_id, position=Position(**position), **kwargs)
        elif rotation is not None:
            super().__init__(object_type="camera", object_id=object_id, rotation=Rotation(**rotation), **kwargs)

    def update_attributes(self, evt_handler=None, **kwargs):
        super().update_attributes(evt_handler=evt_handler, **kwargs)

        self.hasAudio = kwargs.get("hasAudio", False)
        self.hasVideo = kwargs.get("hasVideo", False)
        self.hasAvatar = kwargs.get("hasAvatar", False)
        self.displayName = kwargs.get("displayName", "")
        self.jistsiId = kwargs.get("jistsiId", None)



from ..utils import Utils
from .attribute import Attribute
from .position import Position

class Impulse(Attribute):
    """
    Impulse Attribute.
    Usage: impulse=Impulse(...)
    """
    def __init__(self, on="mousedown", force=Position(0,0,0), position=Position(0,0,0)):
        if isinstance(force, Position):
            force = force.to_str()
        elif isinstance(force, tuple) or isinstance(force, list):
            force = Utils.tuple_to_string(force)

        if isinstance(position, Position):
            position = position.to_str()
        elif isinstance(position, tuple) or isinstance(position, list):
            position = Utils.tuple_to_string(position)

        super().__init__(on=on, force=force, position=position)



"""
auth.py - Authentication methods for accessing the ARENA.
"""

import json
import os
import pickle
import ssl
import sys
import time
import webbrowser
from pathlib import Path
from urllib import parse, request
from urllib.error import HTTPError, URLError
from urllib.parse import urlsplit

import jwt
import requests
from google.auth.transport.requests import AuthorizedSession, Request
from google_auth_oauthlib.flow import InstalledAppFlow

debug_toggle = False
_scopes = ["openid",
           "https://www.googleapis.com/auth/userinfo.profile",
           "https://www.googleapis.com/auth/userinfo.email"]
_user_gauth_path = f'{str(Path.home())}/.arena_google_auth'
_user_mqtt_path = f'{str(Path.home())}/.arena_mqtt_auth'
_local_mqtt_path = f'.arena_mqtt_auth'
_csrftoken = None
_mqtt_token = None
_id_token = None


def authenticate_user(host, debug=False):
    """
    Begins authentication flow, getting Google auth, opening web browser if
    needed, getting username and state from ARENA server.
    host: The hostname of the ARENA webserver.
    debug: True to skip SSL verify for localhost tests.
    Returns: Username from arena-account, or None.
    """
    global debug_toggle
    global _id_token
    debug_toggle = debug
    print("Signing in to the ARENA...")

    local_token = _local_token_check()
    if local_token:
        return local_token["username"]

    creds = None
    browser = None
    try:
        browser = webbrowser.get()
    except (webbrowser.Error) as err:
        print("Console-only login. {0}".format(err))

    # store the user's access and refresh tokens
    if os.path.exists(_user_gauth_path):
        print("Using cached Google authentication.")
        with open(_user_gauth_path, 'rb') as token:
            creds = pickle.load(token)
        session = AuthorizedSession(creds)
    # if no credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            print("Requesting refreshed Google authentication.")
            creds.refresh(Request())
            session = AuthorizedSession(creds)
        else:
            print("Requesting new Google authentication.")
            gauth_json = _get_gauthid(host)
            flow = InstalledAppFlow.from_client_config(
                json.loads(gauth_json), _scopes)
            if browser:
                # TODO: select best client port to avoid likely conflicts
                creds = flow.run_local_server(port=8989)
            else:
                creds = flow.run_console()
            session = flow.authorized_session()
        with open(_user_gauth_path, 'wb') as token:
            # save the credentials for the next run
            pickle.dump(creds, token)
        os.chmod(_user_gauth_path, 0o600)  # set user-only perms.

    username = None
    _id_token = creds.id_token
    user_info = _get_user_state(host, _id_token)
    _user_info = json.loads(user_info)
    if 'authenticated' in _user_info and 'username' in _user_info:
        username = _user_info["username"]
    profile_info = session.get(
        'https://www.googleapis.com/userinfo/v2/me').json()
    if profile_info:
        print(f'Authenticated Google account: {profile_info["email"]}')
    return username


def authenticate_scene(host, realm, scene, username, debug=False):
    """ End authentication flow, requesting permissions may change by owner
    or admin, for now, get a fresh mqtt_token each time.
    host: The hostname of the ARENA webserver.
    realm: The topic realm name.
    scene: The namespace/scene name combination.
    username: The ARENA username for the user.
    debug: True to skip SSL verify for localhost tests.
    Returns: username and mqtt_token from arena-account.
    """
    global debug_toggle
    global _id_token
    global _mqtt_token
    debug_toggle = debug
    local_token = _local_token_check()
    if local_token:
        return local_token["username"]

    print("Using remote-authenticated MQTT token.")
    mqtt_json = _get_mqtt_token(host, realm, scene, username, _id_token)
    # save mqtt_token
    with open(_user_mqtt_path, mode="w") as d:
        d.write(mqtt_json)
    os.chmod(_user_mqtt_path, 0o600)  # set user-only perms.

    _mqtt_token = json.loads(mqtt_json)
    username = None
    if 'username' in _mqtt_token:
        username = _mqtt_token['username']
    print(f'ARENA Username: {username}')
    return _mqtt_token


def _local_token_check():
    # TODO: remove local check after ARTS supports mqtt_token passing
    # check for local mqtt_token first
    if os.path.exists(_local_mqtt_path):
        print("Using local MQTT token.")
        f = open(_local_mqtt_path, "r")
        mqtt_json = f.read()
        f.close()
        # TODO: check token expiration
        _mqtt_token = json.loads(mqtt_json)
        return _mqtt_token
    return None


def _get_csrftoken(host):
    # get the csrftoken for django
    global _csrftoken
    csrf_url = f'https://{host}/user/login'
    client = requests.session()
    if debug_toggle:
        _csrftoken = client.get(csrf_url, verify=False).cookies['csrftoken']
    else:
        _csrftoken = client.get(csrf_url).cookies['csrftoken']
    return _csrftoken


def _get_gauthid(host):
    url = f'https://{host}/conf/gauth.json'
    return urlopen(url)


def _get_user_state(host, id_token):
    global _csrftoken
    url = f'https://{host}/user/user_state'
    if not _csrftoken:
        _csrftoken = _get_csrftoken(host)
    params = {"id_token": id_token}
    query_string = parse.urlencode(params)
    data = query_string.encode("ascii")
    return urlopen(url, data=data, csrf=_csrftoken)


def _get_mqtt_token(host, realm, scene, username, id_token):
    global _csrftoken
    url = f'https://{host}/user/mqtt_auth'
    if not _csrftoken:
        _csrftoken = _get_csrftoken(host)
    params = {
        "id_auth": "google-installed",
        "username": username,
        "id_token": id_token,
        "realm": realm,
        "scene": scene
    }
    query_string = parse.urlencode(params)
    data = query_string.encode("ascii")
    return urlopen(url, data=data, csrf=_csrftoken)


def urlopen(url, data=None, creds=False, csrf=None):
    """ urlopen is for ARENA URL connections.
    url: the url to POST/GET.
    data: None for GET, add params for POST.
    creds: True to pass the MQTT token as a cookie.
    csrf: The csrftoken.
    """
    global debug_toggle
    global _mqtt_token
    try:
        req = request.Request(url)
        if creds:
            req.add_header("Cookie", f"mqtt_token={_mqtt_token['token']}")
        if csrf:
            req.add_header("Cookie", f"csrftoken={csrf}")
            req.add_header("X-CSRFToken", csrf)
        if debug_toggle:
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            res = request.urlopen(req, data=data, context=context)
        else:
            res = request.urlopen(req, data=data)
        return res.read().decode('utf-8')
    except (URLError, HTTPError) as err:
        print("{0}: ".format(err)+url)
        if isinstance(err, HTTPError) and round(err.code, -2) == 400:
            # user not authorized on website yet, they don't have an ARENA username
            base_url = "{0.scheme}://{0.netloc}".format(urlsplit(url))
            print(f'Login with this this account on the website first:')
            print(f'Trying to open login page: {base_url}/user')
            try:
                webbrowser.open_new_tab(f'{base_url}/user')
            except (webbrowser.Error) as err:
                print("Console-only login. {0}".format(err))
        sys.exit("Terminating...")


def signout():
    if os.path.exists(_user_gauth_path):
        os.remove(_user_gauth_path)
    if os.path.exists(_user_mqtt_path):
        os.remove(_user_mqtt_path)
    print("Signed out of the ARENA.")


def _print_mqtt_token(jwt):
    print('ARENA MQTT Permissions')
    print('----------------------')
    print(f'User: {jwt["sub"]}')
    exp_str = time.strftime("%c", time.localtime(jwt["exp"]))
    print(f'Expires: {exp_str}')
    print('Publish topics:')
    for pub in jwt["publ"]:
        print(f'- {pub}')
    print('Subscribe topics:')
    for sub in jwt["subs"]:
        print(f'- {sub}')


def permissions():
    mqtt_path = None
    # check for local mqtt_token first
    if os.path.exists(_local_mqtt_path):
        print("Using local MQTT token.")
        mqtt_path = _local_mqtt_path
    elif os.path.exists(_user_mqtt_path):
        print("Using user MQTT token.")
        mqtt_path = _user_mqtt_path
    if mqtt_path:
        f = open(mqtt_path, "r")
        mqtt_json = f.read()
        f.close()
        mqtt_token = json.loads(mqtt_json)
        decoded = jwt.decode(mqtt_token["token"], options={
            "verify_signature": False})
        _print_mqtt_token(decoded)


if __name__ == '__main__':
    globals()[sys.argv[1]]()
